/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/gpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.pdf.textsegmentation.implementation;

import com.frojasg1.chesspdfbrowser.io.notation.ChessMoveAlgebraicNotation;
import com.frojasg1.chesspdfbrowser.model.chess.configuration.ChessViewConfiguration;
import com.frojasg1.chesspdfbrowser.pdf.renderer.MyPDFRenderer;
import com.frojasg1.chesspdfbrowser.pdf.textsegmentation.PDFPageSegmentatorInterface;
import com.frojasg1.chesspdfbrowser.pdf.textsegmentation.PageSegmentationResult;
import com.frojasg1.general.StringFunctions;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.progress.CancellationException;
import com.frojasg1.general.progress.UpdatingProgress;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.rendering.PDFRenderer;
import org.apache.pdfbox.text.PDFTextStripperByArea;

/**
 *
 * @author Usuario
 */
public class PDFPageSegmentator implements PDFPageSegmentatorInterface
{
	protected Map<Dimension, SummarizedPage> _map = null;

	protected ChessViewConfiguration _configuration = null;

	protected PDFRenderer _pdfRenderer_forBackground = null;
	protected PDFRenderer _pdfRenderer_toSegment = null;
	protected PDDocument _document = null;

	public PDFPageSegmentator( ChessViewConfiguration configuration )
	{
		_configuration = configuration;
		_map = new HashMap<Dimension, SummarizedPage>();
	}

	public SummarizedPage getSummarizedPage( Dimension dim )
	{
		return( _map.get( dim ) );
	}

	public void setChessViewConfiguration( ChessViewConfiguration configuration )
	{
		_configuration = configuration;
	}
	
	public PDDocument getPDDocument()
	{
		return( _document );
	}
	
	public int getNumberOfPages()
	{
		int result = -1;
		
		if( _document != null )
			result = _document.getNumberOfPages();

		return( result );
	}


	protected PDFRenderer createPDFRendererForBackground(PDDocument document)
	{
		MyPDFRenderer mpr = new MyPDFRenderer( document );
		mpr.setShowText(false);

		return( mpr );
	}
	
	protected PDFRenderer createPDFRendererForToSegment( PDDocument document )
	{
		MyPDFRenderer mpr = new MyPDFRenderer( document );

		return( mpr );
	}

	public void initialize(PDDocument document,
							int minimumNumberOfPagesToCheck,
							int maximumNumberOfPagesToCheck,
							UpdatingProgress up )
					throws IOException, CancellationException
	{
		_document = document;
		_pdfRenderer_forBackground = createPDFRendererForBackground(document);
		_pdfRenderer_toSegment = null;

		if( up != null )
			up.up_childStarts();

		for( int ii=0; ii<document.getNumberOfPages(); ii++ )
		{
			PDPage page = document.getPage(ii);

			if( isChessGamePage( page ) )
			{
				Dimension size = new Dimension( (int) page.getBBox().getWidth(),
												(int) page.getBBox().getHeight() );

				SummarizedPage sp = _map.get( size );

				if( sp == null )
				{
					sp = new SummarizedPage();
					sp.setNumberOfPagesToSummarize( minimumNumberOfPagesToCheck, maximumNumberOfPagesToCheck );
					_map.put( size, sp );
				}

				if( ( sp != null ) && !sp.isComplete() )
				{
					BufferedImage image = _pdfRenderer_forBackground.renderImageWithDPI( ii, 72F );

					sp.addPage(image);
				}
			}

			if( up != null )
				up.up_updateProgressFromChild( (double) ii / document.getNumberOfPages() );
		}

		Iterator<Dimension> it = _map.keySet().iterator();
		while( it.hasNext() )
		{
			Dimension key = it.next();
			SummarizedPage sp = _map.get( key );
			sp.summarize();
		}
		if( up != null )
			up.up_childEnds();
	}

	@Override
	public PageSegmentationResult getSegments( int pageNumber ) throws IOException
	{
		if( _pdfRenderer_toSegment == null )
			_pdfRenderer_toSegment = createPDFRendererForToSegment( _document );
		
		PageSegmentationResult result = new PageSegmentationResult();

		PDPage page = _document.getPage( pageNumber );

		Dimension size = new Dimension( (int) page.getBBox().getWidth(),
										(int) page.getBBox().getHeight() );

		SummarizedPage sp = _map.get( size );

		BufferedImage image = _pdfRenderer_toSegment.renderImageWithDPI( pageNumber, 72F );

		if( sp != null )
		{
			ImageSegmentator is = new ImageSegmentator( sp, image, null );
			List<Rectangle> listOfSegments = is.getAreaSegments();

			result.setListOfSegmentedRegions( listOfSegments );
			result.setCouldBeValidated(true);
		}

		return( result );
	}

	@Override
	public String getTextOfPage(int pageNumber) throws IOException
	{
		StringBuilder sb = new StringBuilder();
		
		PageSegmentationResult psr = getSegments( pageNumber );
		
		PDPage page = _document.getPage( pageNumber );
		
		PDFTextStripperByArea stripper = new PDFTextStripperByArea();
		stripper.setSortByPosition( true );

		Iterator<Rectangle> it = psr.getIteratorToSegmentedRegions();
		int index = 0;
		while( it.hasNext() )
		{
			Rectangle rect = it.next();
			stripper.addRegion( "region-" + index, rect );
			index++;
		}

		stripper.extractRegions( page );

		for( int ii=0; ii<index; ii++ )
		{
			String regionText = stripper.getTextForRegion( "region-" + ii );
			sb.append( regionText );
		}

		return( sb.toString() );
	}

	protected boolean isChessGamePage( PDPage page ) throws IOException
	{
		boolean result = false;

		Rectangle rect = new Rectangle( 0, 0,
										(int) page.getBBox().getWidth(),
										(int) page.getBBox().getHeight() );

		PDFTextStripperByArea stripper = new PDFTextStripperByArea();
		stripper.setSortByPosition( true );

		stripper.addRegion( "region", rect );
		stripper.extractRegions( page );
		String text = stripper.getTextForRegion( "region" );

		result = isChessGamePage( text );
		
		return( result );
	}

	protected boolean isChessGamePage( String text )
	{
		LiteParserValidator lpv = new LiteParserValidator( text, _configuration );
		
		return( lpv.isChessGameText() );
	}

	protected static class LiteParserValidator
	{
		protected static final int NUMBER = 1;
		protected static final int DOT = 2;
		protected static final int MOVE = 3;

		protected ChessViewConfiguration _configuration = null;

		enum ResultOfCheck
		{
			MATCHES,
			NOT_COMPLETE,
			FAILS
		}

		protected String _text = null;
		protected int _position = -1;
		List<String> _tokensToCheck = null;

		public LiteParserValidator( String text, ChessViewConfiguration configuration )
		{
			_configuration = configuration;
			_text = text;
			_tokensToCheck = new ArrayList<String>();
		}

		public boolean isChessGameText()
		{
			boolean result = false;

			while( !result && hasNext() )
			{
				String token = next();
				_tokensToCheck.add( token );
				result = checkTokens();
			}

			return( result );
		}
		
		public boolean hasNext()
		{
			return( _position < _text.length() );
		}

		public String next()
		{
			String result = "";

			while( ( result.length() == 0 ) && hasNext() )
			{
				if( _position >= 0 )
				{
					String separator =  _text.substring( _position, _position + 1 );
					if( separator.equals( "." ) )
						_tokensToCheck.add( separator );
				}

				int pos1 = StringFunctions.indexOfAnyChar( _text, " \t.", _position + 1 );
				if( pos1 == -1 )		pos1 = _text.length();

				result = _text.substring( _position + 1, pos1 );
				_position = pos1;
			}

			return( result );
		}
		
		protected boolean checkTokens()
		{
			boolean result = false;

			final int[] sequence1 = new int[] { NUMBER, DOT, MOVE, MOVE };
			final int[] sequence2 = new int[] { NUMBER, DOT, DOT, DOT, MOVE, NUMBER, DOT, MOVE };

			ResultOfCheck result1 = check( sequence1 );
			ResultOfCheck result2 = null;
			if( ! result1.equals( ResultOfCheck.MATCHES ) )
			{
				result2 = check( sequence2 );
			}

			if( result1.equals( ResultOfCheck.MATCHES ) || result2.equals( ResultOfCheck.MATCHES ) )
				result = true;
			else if( !result1.equals( ResultOfCheck.NOT_COMPLETE ) && !result2.equals( ResultOfCheck.NOT_COMPLETE ) )
			{
				_tokensToCheck.clear();
			}

			return( result );
		}

		protected boolean isItAChessMoveString( String moveString )
		{
			boolean result = false;
			String englishMoveString = _configuration.getChessLanguageConfigurationToParseTextFrom().translateMoveStringToEnglish( moveString );
			result = ChessMoveAlgebraicNotation.getInstance().isItAChessMoveString( englishMoveString );
			return( result );
		}

		protected ResultOfCheck check( int[] sequence )
		{
			ResultOfCheck result = ResultOfCheck.NOT_COMPLETE;

			int lastNumber = -1;
			Iterator<String> it = _tokensToCheck.iterator();
			for( int ii=0; it.hasNext() && (ii<sequence.length) && result.equals( ResultOfCheck.NOT_COMPLETE ); ii++ )
			{
				String token = it.next();
				switch( sequence[ii] )
				{
					case NUMBER:
					{
						Integer number = IntegerFunctions.parseInt(token);
						if( ( number == null ) ||
							( lastNumber != -1 ) && ( ( lastNumber + 1 ) != number ) )
						{
							result = ResultOfCheck.FAILS;
						}
						else
						{
							lastNumber = number;
						}
					}
					break;

					case DOT:
						if( ! token.equals( "." ) )
							result = ResultOfCheck.FAILS;
					break;

					case MOVE:
						if( ! isItAChessMoveString( token ) )
							result = ResultOfCheck.FAILS;
					break;
				}
				if( (ii==(sequence.length-1)) && result.equals( ResultOfCheck.NOT_COMPLETE ) )
					result = ResultOfCheck.MATCHES;
			}
			
			return( result );
		}
	}
}
