/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.view.chess;

import com.frojasg1.chesspdfbrowser.configuration.ChessStringsConf;
import com.frojasg1.chesspdfbrowser.model.chess.board.ChessBoard;
import com.frojasg1.chesspdfbrowser.model.chess.board.ChessGameResult;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.ChessGameMove;
import com.frojasg1.chesspdfbrowser.model.chess.pieces.ChessPiece;
import com.frojasg1.chesspdfbrowser.view.chess.images.ChessBoardImages;
import com.frojasg1.chesspdfbrowser.view.chess.interaction.ChessGameControllerInterface;
import com.frojasg1.chesspdfbrowser.view.chess.interaction.ChessGameViewInterface;
import com.frojasg1.applications.common.configuration.application.ChangeLanguageClientInterface;
import com.frojasg1.applications.common.configuration.application.ChangeLanguageServerInterface;
import com.frojasg1.general.desktop.ImageFunctions;
import com.frojasg1.general.desktop.mouse.AvailableCursors;
import com.frojasg1.general.number.IntegerFunctions;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.util.List;
import java.awt.Rectangle;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.Iterator;
import javax.swing.JPanel;
import javax.swing.JSplitPane;

/**
 *
 * @author Usuario
 */
public class ChessBoardPanel extends JPanel implements ComponentListener, MouseListener,
														MouseMotionListener, ChessGameViewInterface,
														PromotionControllerInterface, ChangeLanguageClientInterface
{
	protected ChangeLanguageServerInterface _changeLanguageServer = null;

	protected ChessGameControllerInterface _listener = null;

	protected ChessBoard _chessBoard = null;
	protected boolean _whitePlaysFromTheBottom = true;

	protected double _fractionOfPieceWidthOverSquareWidth = 0.95d;
	
	protected int _currentSquareWidth = -1;
	protected int _currentPieceWidth = -1;
	protected int _initialX_ofBoard = -1;
	
	protected static final int MIN_SQUARE_WIDTH = 32;

	protected ChessBoardImages _chessBoardImages = null;

	protected String _pieceCodeToShow = null;
	protected boolean _hasToShowPicture = false;

	protected int _currentCursorColumn = -1;
	protected int _currentCursorRow = -1;
	
	protected boolean _inDrag = false;
	protected int _startDragColumn = -1;
	protected int _startDragRow = -1;

	protected boolean _lastIsInPositionToDrag = false;
	protected boolean _lastMouseIsOverAPiece = false;
	
	protected int _currentX = -1;
	protected int _currentY = -1;
	
//	protected static final Cursor _handCursor = new Cursor( Cursor.HAND_CURSOR );
//	protected static final Cursor _waitCursor = new Cursor( Cursor.WAIT_CURSOR );
//	protected static final Cursor _defaultCursor = new Cursor( Cursor.DEFAULT_CURSOR );

	protected PromotionJPanel _promotionJPanel = null;
	protected ChessGameMove _promotionMove = null;
	
	public ChessBoardPanel() throws IOException
	{
		super();

		Dimension dim = new Dimension(	MIN_SQUARE_WIDTH * (ChessBoard.NUM_OF_COLUMNS+1),
										MIN_SQUARE_WIDTH * (ChessBoard.NUM_OF_ROWS+1));
		this.setMinimumSize(dim);

		setLayout(null);
		setOpaque( false );

		_chessBoardImages = ChessBoardImages.instance();

		_promotionJPanel = new PromotionJPanel( this );
		_promotionJPanel.setVisible( false );

		add( _promotionJPanel );

//		_promotionJPanel.setBounds( 0, 0, 160, 40 );

		registerToChangeLanguageAsObserver( ChessStringsConf.instance() );

		addComponentListener(this);
		addMouseMotionListener(this);
		addMouseListener(this);
	}

	public void setListener( ChessGameControllerInterface listener )
	{
		_listener = listener;
	}

	public void setShowPicture( String pieceCode )
	{
		_pieceCodeToShow = pieceCode;
		_hasToShowPicture = true;
		repaint();
	}

	@Override
	public void setChessBoard( ChessBoard cb )
	{
		synchronized( this )
		{
			_chessBoard = cb;
		}
	}

	public ChessBoard getChessBoard()
	{
		synchronized( this )
		{
			return( _chessBoard );
		}
	}

	public int getCurrentSquareWidth()
	{
		return( _currentSquareWidth );
	}
	
	public int getCurrentPieceWidth()
	{
		return( _currentPieceWidth );
	}
	
	public void setWhitePlaysFromTheBottom( boolean whitePlaysInTheBottom )
	{
		_whitePlaysFromTheBottom = whitePlaysInTheBottom;

		repaint();
	}

	public boolean getWhitePlaysInTheBottom()
	{
		return( _whitePlaysFromTheBottom );
	}

	protected boolean isLegalMovePromotion( ChessGameMove cgm )
	{
		boolean result = false;
		
		if( _chessBoard != null )
		{
			ChessPiece piece = _chessBoard.getPiece_fast( cgm._column1, cgm._row1 );
			if( piece != null )
			{
				result = piece.legalMoveIsPromotion(cgm);
			}
		}

		return( result );
	}
	
	protected void makeVisiblePromotionPanel( Point location )
	{
		Point locationOfQueenButton = _promotionJPanel.getCenterPointOfQueenButton();

		int xx = (int) (location.getX() - locationOfQueenButton.getX());
		int yy = (int) (location.getY() - locationOfQueenButton.getY());

		xx = IntegerFunctions.min( IntegerFunctions.max( 0, xx ), getWidth() - _promotionJPanel.getWidth() );
		yy = IntegerFunctions.min( IntegerFunctions.max( 0, yy ), getHeight() - _promotionJPanel.getHeight() );

		_promotionJPanel.setBounds( xx, yy, (int) _promotionJPanel.getSize().getWidth(), (int) _promotionJPanel.getSize().getHeight() );
		_promotionJPanel.setColor( _chessBoard.getColorToPlay() );

		_promotionJPanel.setVisible( true );
	}

	protected void makeNewMove( ChessGameMove cgm )
	{
		if( _chessBoard != null )
		{
			List<ChessGameMove> list = _chessBoard.getListOfMoves();
			list.add(cgm);
			_listener.newPositionInTheMovesTree(list);
		}
	}
	
	
	@Override
	public void componentHidden(ComponentEvent e)
	{
	}

	@Override
	public void componentMoved(ComponentEvent e)
	{
	}

	@Override
	public void componentResized(ComponentEvent e)
	{
		if( e.getComponent() == this )
		{
			changeCurrentSquareWidth();

			repaint();
		}
	}

	// Five methods from MouseListener:
	/** Called when the mouse has been clicked on a component. */
	@Override
	public void mouseClicked(MouseEvent e)
	{
	}

	/** Called when the mouse enters a component. */
	@Override
	public void mouseEntered(MouseEvent e)
	{
		_inDrag = false;
	}

	/** Called when the mouse exits a component. */
	@Override
	public void mouseExited(MouseEvent e)
	{
		if( _inDrag )
		{
			_currentCursorColumn = -1;
			_currentCursorRow = -1;
			_inDrag = false;
			repaint();
		}
	}

	/** Called when the mouse has been pressed. */
	@Override
	public void mousePressed(MouseEvent e)
	{	
		if( ! getCursor().getName().equals( AvailableCursors._waitCursor.getName() ) )	// if we are loading or saving, we must not set the hand cursor.
		{
/*			if( ( e.getButton() == MouseEvent.BUTTON2 ) || ( e.getButton() == MouseEvent.BUTTON3 ) )
			{
				int color = -1;
				if( e.getButton() == MouseEvent.BUTTON2 )		color = ChessPiece.WHITE;
				else if( e.getButton() == MouseEvent.BUTTON3 )	color = ChessPiece.BLACK;

				_promotionJPanel.setColor( color );
				_promotionJPanel.setBounds( e.getX(), e.getY(), (int) _promotionJPanel.getSize().getWidth(), (int) _promotionJPanel.getSize().getHeight() );
				_promotionJPanel.setVisible(true);
				repaint();
			}
			else */
			if( e.getButton() == MouseEvent.BUTTON1 ) 
			{
				calculateCurrentCursorColAndRow( e.getX(), e.getY() );
				if( isInPositionToDrag( e.getX(), e.getY(), true ) != null )
				{
					_currentX = e.getX();
					_currentY = e.getY();
					_startDragColumn = _currentCursorColumn;
					_startDragRow = _currentCursorRow;
					_inDrag = true;

					repaint();
				}
			}
		}
	}

	/** Called when the mouse has been released. */
	@Override
	public void mouseReleased(MouseEvent e)
	{
		if( ! getCursor().getName().equals( AvailableCursors._waitCursor.getName() ) )	// if we are loading or saving, we must not set the hand cursor.
		{
/*			if( ( e.getButton() == MouseEvent.BUTTON2 ) || ( e.getButton() == MouseEvent.BUTTON3 ) )
				_promotionJPanel.setVisible(false);
			else */
			if( e.getButton() == MouseEvent.BUTTON1 )
			{
				_inDrag = false;

				calculateCurrentCursorColAndRow( e.getX(), e.getY() );

				ChessGameMove cgm = new ChessGameMove( _startDragColumn, _startDragRow, _currentCursorColumn, _currentCursorRow );

				try
				{
					cgm.checkIsMoveInsideBoard();  // throws exception if not

					if( (_listener != null) && ( _chessBoard != null ) && _chessBoard.isLegalThisMove( cgm, false ) )
					{
						if( isLegalMovePromotion( cgm ) )
						{
							makeVisiblePromotionPanel( e.getPoint() );
							_promotionMove = cgm;
						}
						else
						{
							makeNewMove( cgm );
						}
					}
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}

				repaint();
			}
		}
	}

	@Override
	public void componentShown(ComponentEvent e)
	{
		if( e.getComponent() == this )
		{
			changeCurrentSquareWidth();
		}
	}

	@Override
	public void mouseMoved( MouseEvent evt )
	{
		if( ! getCursor().getName().equals( AvailableCursors._waitCursor.getName() ) )	// if we are loading or saving, we must not set the hand cursor.
		{
			if( evt.getComponent() == this )
			{
				boolean hasChanged = calculateCurrentCursorColAndRow( evt.getX(), evt.getY() );
				if( setAppropriateCursor( evt, hasChanged ) || hasChanged )
				{
					repaint();
				}
			}
		}
		else
		{
			
		}
	}

	@Override
	public void mouseDragged( MouseEvent evt )
	{
		if( ! getCursor().getName().equals( AvailableCursors._waitCursor.getName() ) )	// if we are loading or saving, we must not set the hand cursor.
		{
			_currentX = evt.getX();
			_currentY = evt.getY();

			calculateCurrentCursorColAndRow( _currentX, _currentY );
			if( _inDrag )
			{
				repaint();
			}
		}
	}

	public boolean calculateCurrentCursorColAndRow( int currentX, int currentY )
	{
		int previousCursorColumn = _currentCursorColumn;
		int previousCursorRow = _currentCursorRow;
		
		_currentCursorColumn = -1;
		_currentCursorRow = -1;

		int xx = currentX - _initialX_ofBoard + _currentSquareWidth;
		int yy = currentY - _initialX_ofBoard + _currentSquareWidth;

		int column = xx/_currentSquareWidth - 1;
		int row = yy/_currentSquareWidth - 1;

		if( isPositionInsideTheBoard( column+1, row+1 ) )
		{
			if( _whitePlaysFromTheBottom )
			{
				_currentCursorColumn = column+1;
				_currentCursorRow = ChessBoard.NUM_OF_ROWS - row;
			}
			else
			{
				_currentCursorColumn = ChessBoard.NUM_OF_COLUMNS - column;
				_currentCursorRow = row+1;
			}
		}
		
		boolean result =	( previousCursorColumn != _currentCursorColumn ) ||
							( previousCursorRow != _currentCursorRow );

		return( result );
	}

	protected ChessPiece isInPositionToDrag( int xx, int yy, boolean hasChangedSquare )
	{
		ChessPiece result = null;
		
		xx = xx - _initialX_ofBoard;
		yy = yy - _initialX_ofBoard;

		if( isPositionInsideTheBoard( _currentCursorColumn, _currentCursorRow ) )
		{
			if( ( _chessBoard != null ) && (_chessBoard.getChessGameResult() == ChessGameResult.GAME_CONTINUES ) )
			{
				ChessPiece cp = null;
				boolean canMove = false;

				try
				{
					cp = _chessBoard.getPiece( _currentCursorColumn, _currentCursorRow );
					canMove = (cp != null) && (cp.getColor() == _chessBoard.getColorToPlay() );
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}

				int xInsideSquare = ( xx % _currentSquareWidth );
				int yInsideSquare = ( yy % _currentSquareWidth );

				if( ( xInsideSquare >= (0.15d*_currentSquareWidth) ) &&
					( xInsideSquare <= (0.85d*_currentSquareWidth) ) &&
					( yInsideSquare >= (0.15d*_currentSquareWidth) ) &&
					( yInsideSquare <= (0.85d*_currentSquareWidth) ) )
				{
					if( _lastMouseIsOverAPiece && !hasChangedSquare )
					{
						if( _lastIsInPositionToDrag )
							result = cp;
					}
					else if( canMove && ( cp.getListOfLegalMoves().size() > 0 ) )
					{
						result = cp;
						_lastIsInPositionToDrag = true;
					}
					else
					{
						_lastIsInPositionToDrag = false;
					}
				}
				else
				{
					_lastMouseIsOverAPiece = false;
				}
			}
		}

		return( result );
	}

	public boolean setAppropriateCursor( MouseEvent evt, boolean hasChangedSquare )
	{
		boolean hasChanged = false;

		if( ! getCursor().getName().equals( AvailableCursors._waitCursor.getName() ) )	// if we are loading or saving, we must not set the hand cursor.
		{
			Cursor cursor = AvailableCursors._defaultCursor;

	//		calculateCurrentCursorColAndRow( evt.getX(), evt.getY() );

			if( isInPositionToDrag( evt.getX(), evt.getY(), hasChangedSquare ) != null )
			{
				cursor = AvailableCursors._handCursor;
			}

			hasChanged = !getCursor().getName().equals( cursor.getName() );
			setCursor( cursor );
		}

		return( hasChanged );
	}

	protected String getDescriptionParentSize( Component comp, int depth )
	{
		String result = "null";
		Component parent = comp;
		for( int ii=0; (parent!=null) && (ii<depth); ii++ )
		{
			parent = parent.getParent();
		}

		if( parent != null )
		{
			String dividerLocationStr = ( parent instanceof JSplitPane ? "-dl(" + ((JSplitPane)parent).getDividerLocation() + ")" : "" );
			result = parent.getClass().getName() + "-" + parent.getName() +
					"(" + ((int)parent.getSize().getWidth()) + "," +
					((int)parent.getSize().getHeight()) + ")" +
					dividerLocationStr;
		}

		return( result );
	}

	@Override
	public void paintComponent(Graphics gc)
	{
		synchronized( this )
		{
//			super.paintComponent( gc );
/*
		System.out.println( "Size: " + getDescriptionParentSize( this, 0 ) +
							". Parent Size: " + getDescriptionParentSize( this, 1 ) +
							". Parent.Parent Size: " + getDescriptionParentSize( this, 2 ) +
							". Parent.Parent.Parent Size: " + getDescriptionParentSize( this, 3 ) +
							". Parent.Parent.Parent.Parent Size: " + getDescriptionParentSize( this, 4 )
							);
*/

			if( ! _hasToShowPicture )
				paintChessBoardPosition( gc );
			else
				showPicture( gc );

			super.paintComponent( gc );

			_hasToShowPicture = false;
		}
	}

	public void showPicture( Graphics gc )
	{
		
		BufferedImage bi = null;
		if( _pieceCodeToShow.equals( "WS" ) )
			bi = _chessBoardImages.getSquareImage( ChessPiece.WHITE, _currentSquareWidth );
		else if( _pieceCodeToShow.equals( "BS" ) )
			bi = _chessBoardImages.getSquareImage( ChessPiece.BLACK, _currentSquareWidth );
		else
			bi = _chessBoardImages.getPieceImage(_pieceCodeToShow, _currentPieceWidth );

		if( bi != null )
			gc.drawImage( bi, 0, 0, bi.getWidth(), bi.getHeight(), 0, 0, bi.getWidth(), bi.getHeight(), null );
		else
		{
			gc.setColor( Color.BLACK );
			gc.drawString( "null", 30, 30 );
		}
	}

	protected void paintChessBoardPosition( Graphics gc )
	{
		BufferedImage bi = new BufferedImage( getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB  );
		Graphics gc1 = bi.getGraphics();

		Color color = new Color( 0xFFb0b0b0 );
		gc1.setColor(color);
		gc1.fillRect( 0, 0, getWidth(), getHeight() );

		paintBoard( gc1 );
		writeColumnLettersAndRowNumbers( gc1 );

		if( ( _chessBoard != null ) && ! _chessBoard.isCurrentNodeIllegal() )
		{
			markOriginOfMoveAndPossibleDestinations( gc1 );
			paintTurn( gc1 );
			paintResult( gc1 );

			showColAndRow( gc1 );

			paintPieces( gc1 );
		}
		else if( ( _chessBoard != null ) && _chessBoard.isCurrentNodeIllegal() )
		{
			writeIllegalPosition( gc1 );
		}
		else
		{
			writeInitialPositionNotSet( gc1 );
		}

		gc.drawImage( bi, 0, 0, bi.getWidth(), bi.getHeight(), 0, 0, bi.getWidth(), bi.getHeight(), null );
	}

	protected void showColAndRow( Graphics gc )
	{
//		gc.drawString( _currentCursorColumn + ", " + _currentCursorRow, 10, 25 );
	}

	
	protected void putSquaredFrame( Graphics gc, int col, int row, Color color )
	{
		gc.setColor( color );
		
		Rectangle bounds = getRectangleBoundsOfSquareByColRow( col, row );
		
		gc.drawRect( (int) bounds.getX(), (int) bounds.getY(), (int) bounds.getWidth(), (int) bounds.getHeight() );
		gc.drawRect( (int) bounds.getX()+1, (int) bounds.getY()+1, (int) bounds.getWidth()-2, (int) bounds.getHeight()-2 );
		gc.drawRect( (int) bounds.getX()+2, (int) bounds.getY()+2, (int) bounds.getWidth()-4, (int) bounds.getHeight()-4 );
	}
	
	protected boolean isPositionInsideTheBoard( int column, int row )
	{
		boolean result = ( ( column <= ChessBoard.NUM_OF_COLUMNS ) && ( column > 0 ) &&
							( row <= ChessBoard.NUM_OF_ROWS ) && ( row > 0 ) );
		return( result );
	}

	protected void markOriginOfMoveAndPossibleDestinations( Graphics gc )
	{
		if( _chessBoard != null )
		{
			ChessGameMove cgm = _chessBoard.peekLastMove();

			if( cgm != null )
			{
				putSquaredFrame( gc, cgm._column1, cgm._row1, Color.ORANGE );
				putSquaredFrame( gc, cgm._column2, cgm._row2, Color.ORANGE );
			}

			try
			{
				if( _inDrag )
				{
					putSquaredFrame( gc, _startDragColumn, _startDragRow, Color.YELLOW );
					cgm = new ChessGameMove( _startDragColumn, _startDragRow, _currentCursorColumn, _currentCursorRow );
					if( _chessBoard.isLegalThisMove(cgm, false) )
					{
						putSquaredFrame( gc, _currentCursorColumn, _currentCursorRow, Color.GREEN.darker() );
					}
				}
				else if( (getCursor().equals(AvailableCursors._handCursor)) &&
							isPositionInsideTheBoard( _currentCursorColumn, _currentCursorRow ) )
				{
					ChessPiece cp = _chessBoard.getPiece( _currentCursorColumn, _currentCursorRow );

					if( ( cp != null ) && ( cp.getColor() == _chessBoard.getColorToPlay() ) )
					{
						List<ChessGameMove> listOfMoves = cp.getListOfLegalMoves();
						if( listOfMoves.size() > 0 )
						{
							putSquaredFrame( gc, cp.getColumn(), cp.getRow(), Color.YELLOW );

							Iterator<ChessGameMove> it = listOfMoves.iterator();

							while( it.hasNext() )
							{
								cgm = it.next();
								putSquaredFrame( gc, cgm._column2, cgm._row2, Color.GREEN.darker() );
							}
						}
					}
				}
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}
	}
	
	protected void paintTurn( Graphics gc )
	{
		if( ( _chessBoard != null ) && ( _chessBoard.getCurrentPosition() != null ) )
		{
			if( _chessBoard.getChessGameResult() == ChessGameResult.GAME_CONTINUES )
			{
				Color rectangleColor = null;
				Color boundColor = null;
				int row = 0;

				int y_increment = 0;
				int x_increment = 0;

				if( _whitePlaysFromTheBottom )	x_increment = 4;
				else							x_increment = 2;

				if( _chessBoard.getCurrentPosition().getIsWhitesTurn() )
				{
					rectangleColor = Color.WHITE;
					boundColor = Color.BLACK;
					row = 1;

					if( _whitePlaysFromTheBottom ) y_increment = _initialX_ofBoard + 2;
				}
				else
				{
					rectangleColor = Color.BLACK;
					boundColor = Color.WHITE;
					row = 8;

					if( !_whitePlaysFromTheBottom ) y_increment = _initialX_ofBoard + 1;
				}

				Rectangle squareBound = getRectangleBoundsOfSquareByColRow( 9, row );

				gc.setColor( boundColor );
				gc.drawRect(	(int) ( squareBound.getX() + x_increment ),
								(int) ( squareBound.getY() + y_increment + 2 ),
								(int) ( _initialX_ofBoard - 7 ),
								(int) ( _initialX_ofBoard - 7 ));

				gc.setColor( rectangleColor );
				gc.fillRect(	(int) ( squareBound.getX() + x_increment + 1 ),
								(int) ( squareBound.getY() + y_increment + 3 ),
								(int) ( _initialX_ofBoard - 9 ),
								(int) ( _initialX_ofBoard - 9 ));
			}
		}
	}

	protected void paintResult( Graphics gc )
	{
		String status = null;
		
		if( ( _chessBoard != null ) && ( _chessBoard.getCurrentPosition() != null ) )
		{
			if( _chessBoard.getChessGameResult() == ChessGameResult.GAME_CONTINUES )
			{
				if( _chessBoard.getCurrentPosition().getIsWhitesTurn() )
					status = getChessStrConf().getProperty( ChessStringsConf.CONF_WHITE_TO_PLAY );
				else
					status = getChessStrConf().getProperty( ChessStringsConf.CONF_BLACK_TO_PLAY );

				if( _chessBoard.isCheck() )
					status = getChessStrConf().getProperty( ChessStringsConf.CONF_CHECK ) + " " + status;
			}
			else if( _chessBoard.getChessGameResult() == ChessGameResult.WHITE_WINS )
				status = "1-0 " + getChessStrConf().getProperty( ChessStringsConf.CONF_BLACK_RESIGNS );
			else if( _chessBoard.getChessGameResult() == ChessGameResult.BLACK_WINS )
				status = "0-1" + getChessStrConf().getProperty( ChessStringsConf.CONF_WHITE_RESINGS );
			else if( _chessBoard.getChessGameResult() == ChessGameResult.WHITE_WINS_CHECK_MATE )
				status = getChessStrConf().getProperty( ChessStringsConf.CONF_CHECK_MATE ) + " 1-0 " +
					getChessStrConf().getProperty( ChessStringsConf.CONF_WHITE_WINS );
			else if( _chessBoard.getChessGameResult() == ChessGameResult.BLACK_WINS_CHECK_MATE )
				status = getChessStrConf().getProperty( ChessStringsConf.CONF_CHECK_MATE ) + " 0-1 " +
					getChessStrConf().getProperty( ChessStringsConf.CONF_BLACK_WINS );
			else
			{
				status = getChessStrConf().getProperty( ChessStringsConf.CONF_DRAW ) + "  1/2-1/2";

				if( _chessBoard.getChessGameResult() == ChessGameResult.DRAW_STALE_MATE )
					status = status + "  " + getChessStrConf().getProperty( ChessStringsConf.CONF_STALE_MATE );
				else if( _chessBoard.getChessGameResult() == ChessGameResult.DRAW_THIRD_REPETITION )
					status = status + "  " + getChessStrConf().getProperty( ChessStringsConf.CONF_THIRD_REPETITION );
				else if( _chessBoard.getChessGameResult() == ChessGameResult.DRAW_FIFTY_MOVES_WITHOUT_PROGRESS )
					status = status + getChessStrConf().getProperty( ChessStringsConf.CONF_FIFTY_MOVES_WITHOUT_PROGRESS );
				else if( _chessBoard.getChessGameResult() == ChessGameResult.DRAW_MUTUAL_AGREEMENT )
					status = status + "  " + getChessStrConf().getProperty( ChessStringsConf.CONF_BY_MUTUAL_AGREEMENT );
			}

			Rectangle rect = getRectangleBoundsOfSquareByColRow( 0, 9 );

			Rectangle statusRectangle = new Rectangle(	_initialX_ofBoard,
														(int) rect.getY() + 4,
														_currentSquareWidth * ChessBoard.NUM_OF_COLUMNS,
														(int) rect.getHeight() - 8 );

			int fontSize = (int) Math.floor( _currentSquareWidth * 0.25d );
			fontSize = IntegerFunctions.max( fontSize, 6 );
			fontSize = IntegerFunctions.min( fontSize, 25 );

			Font font = new Font( "Lucida", Font.BOLD, fontSize );
			gc.setFont(font);

			Color brightGreen = new Color( 227, 255, 221, 255 );

			ImageFunctions.paintStringCentered( gc, font, status, Color.BLACK, statusRectangle, brightGreen );
		}
	}
/*
	protected void paintPiece( Graphics gc, ChessPiece cp, Rectangle bounds )
	{
		if( cp != null )
		{
			// if some part of the piece is inside the visible zone (the screen)
			if( ( bounds.getX() < getWidth() ) && ( bounds.getX() + bounds.getWidth() >= 0 ) &&
				( bounds.getY() < getHeight() ) && ( bounds.getY() + bounds.getHeight() >= 0 ) )
			{
				int dx1 = IntegerFunctions.max( 0, (int) bounds.getX() );
				int dy1 = IntegerFunctions.max( 0, (int) bounds.getY() );
				int dx2 = IntegerFunctions.min( getWidth(), (int) ( bounds.getX() + bounds.getWidth() ) );
				int dy2 = IntegerFunctions.min( getHeight(), (int) ( bounds.getY() + bounds.getHeight() ) );

				int sx1 = dx1 - (int) bounds.getX();
				int sy1 = dy1 - (int) bounds.getY();
				int sx2 = sx1 + dx2 - dx1;
				int sy2 = sy1 + dy2 - dy1;

				BufferedImage bi = _chessBoardImages.getPieceImage( cp.getPieceCode() );

				gc.drawImage( bi, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, null );
			}
		}
	}
*/
	protected void paintPieces( Graphics gc )
	{
		if( _chessBoard != null )
		{
			for( int col=1; col<=ChessBoard.NUM_OF_COLUMNS; col++ )
				for( int row=1; row<=ChessBoard.NUM_OF_ROWS; row++ )
				{
					if( ! _inDrag || (col != _startDragColumn) || (row != _startDragRow) )
					{
						try
						{
							ChessPiece cp = _chessBoard.getPiece(col, row);

							if( ( cp != null ) &&
								( ( _promotionMove == null ) ||
								  ( cp.getColumn() != _promotionMove._column1 ) ||
								  ( cp.getRow() != _promotionMove._row1 ) ) )
							{
								Point upperLeftCorner = getUpperLeftCornerOfPieceByColRow( col, row );
								BufferedImage bi = _chessBoardImages.getPieceImage( cp.getPieceCode(), _currentPieceWidth );
								ImageFunctions.paintClippedImage(this, gc, bi, upperLeftCorner);
//								paintPiece( gc, cp, bounds );
							}
						}
						catch( Throwable th )
						{
							th.printStackTrace();
						}
					}
				}

			paintPieceInDrag( gc );
		}
	}

	protected void paintPieceInDrag( Graphics gc )
	{
		if( _inDrag )
		{
			try
			{
				ChessPiece cp = _chessBoard.getPiece( _startDragColumn, _startDragRow );
				if( ( cp != null ) &&
					( ( _promotionMove == null ) ||
					  ( cp.getColumn() != _promotionMove._column1 ) ||
					  ( cp.getRow() != _promotionMove._row1 ) ) )
				{
					Point upperLeftCorner = new Point(	_currentX - _currentPieceWidth/2,
														_currentY - _currentPieceWidth/2 );
					BufferedImage bi = _chessBoardImages.getPieceImage( cp.getPieceCode(), _currentPieceWidth );
					ImageFunctions.paintClippedImage(this, gc, bi, upperLeftCorner);
//					paintPiece( gc, cp, bounds );
				}
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}
	}
	
	protected void paintBoard( Graphics gc )
	{
		for( int col=1; col<=ChessBoard.NUM_OF_COLUMNS; col++ )
			for( int row=1; row<=ChessBoard.NUM_OF_ROWS; row++ )
			{
				Rectangle bounds = getRectangleBoundsOfSquareByColRow( col, row );
				BufferedImage bi = _chessBoardImages.getSquareImage( (col + row)%2, _currentSquareWidth );

				if( bi != null )
				{
					gc.drawImage( bi, (int)bounds.getX(), (int)bounds.getY(),
										(int)(bounds.getX()+bounds.getWidth()),
										(int)(bounds.getHeight()+bounds.getY()),
									0, 0, bi.getWidth(), bi.getHeight(), null );
				}
			}

		gc.setColor( Color.BLACK );

		int x1 = _initialX_ofBoard;
		int y1 = _initialX_ofBoard;
		int deltaX = ChessBoard.NUM_OF_COLUMNS*_currentSquareWidth-1;
		int deltaY = ChessBoard.NUM_OF_ROWS*_currentSquareWidth-1;
		gc.drawRect( x1, y1, deltaX, deltaY);
		gc.drawRect( x1-1, y1-1, deltaX+2, deltaY+2);
		gc.drawRect( x1-2, y1-2, deltaX+4, deltaY+4);
		gc.drawRect( x1-3, y1-3, deltaX+6, deltaY+6);
	}

	protected void writeColumnLettersAndRowNumbers( Graphics gc )
	{
		final String COLUMN_NAMES="abcdefgh";

		int fontSize = (int) Math.floor( _currentSquareWidth * 0.33d );
		fontSize = IntegerFunctions.max( fontSize, 6 );
		fontSize = IntegerFunctions.min( fontSize, 25 );

		Font font = new Font( "Lucida", Font.BOLD, fontSize );
//		gc.setColor( Color.WHITE );
		gc.setFont(font);

		for( int ii=1; ii<9; ii++ )
		{
			String columnName = COLUMN_NAMES.substring( ii-1, ii );

			Rectangle rect = getRectangleBoundsOfSquareByColRow( ii, 0 );
			ImageFunctions.paintStringCentered( gc, font, columnName, Color.WHITE, rect, null );

			rect = getRectangleBoundsOfSquareByColRow( 0, ii );
			ImageFunctions.paintStringCentered( gc, font, String.valueOf(ii), Color.WHITE, rect, null );
		}
	}

	protected void writeInitialPositionNotSet( Graphics gc )
	{
		final String TEXT[]= new String[] {
			getChessStrConf().getProperty( ChessStringsConf.CONF_INITIAL_POSITION_NOT_SET_1 ),
			getChessStrConf().getProperty( ChessStringsConf.CONF_INITIAL_POSITION_NOT_SET_2 ),
			getChessStrConf().getProperty( ChessStringsConf.CONF_INITIAL_POSITION_NOT_SET_3 ) };

		int fontSize = (int) Math.floor( _currentSquareWidth * 0.66d );
		fontSize = IntegerFunctions.max( fontSize, 16 );
		fontSize = IntegerFunctions.min( fontSize, 100 );

		Font font = new Font( "Lucida", Font.BOLD, fontSize );
//		gc.setColor( Color.RED );
		gc.setFont(font);

		int initialX = _initialX_ofBoard;
		int width = 8 * _currentSquareWidth;
		
		int initialY = 3 * _currentSquareWidth;
		int yIncrement = _currentSquareWidth;

		int currentY = initialY;
		for( int ii=0; ii<TEXT.length; ii++ )
		{
			Rectangle rect = new Rectangle( initialX, currentY, width, yIncrement );
			ImageFunctions.paintStringCentered( gc, font, TEXT[ii], Color.RED, rect, null );
			currentY += yIncrement;
		}
	}

	protected void writeIllegalPosition( Graphics gc )
	{
		final String TEXT[]= new String[] {
			getChessStrConf().getProperty( ChessStringsConf.CONF_ILLEGAL_POSITION_1 ),
			getChessStrConf().getProperty( ChessStringsConf.CONF_ILLEGAL_POSITION_2 ) };

		int fontSize = (int) Math.floor( _currentSquareWidth * 0.66d );
		fontSize = IntegerFunctions.max( fontSize, 16 );
		fontSize = IntegerFunctions.min( fontSize, 100 );

		Font font = new Font( "Lucida", Font.BOLD, fontSize );
//		gc.setColor( Color.RED );
		gc.setFont(font);

		int initialX = _initialX_ofBoard;
		int width = 8 * _currentSquareWidth;
		
		int initialY = 3 * _currentSquareWidth;
		int yIncrement = _currentSquareWidth;

		int currentY = initialY;
		for( int ii=0; ii<TEXT.length; ii++ )
		{
			Rectangle rect = new Rectangle( initialX, currentY, width, yIncrement );
			ImageFunctions.paintStringCentered( gc, font, TEXT[ii], Color.RED, rect, null );
			currentY += yIncrement;
		}
	}


	protected Point getUpperLeftCornerOfPieceByColRow( int col, int row )
	{
		Rectangle rect = getRectangleBoundsOfSquareByColRow( col, row );
		
		int x1 = (int) Math.floor( rect.getX() + ( rect.getWidth() - _currentPieceWidth ) / 2 );
		int y1 = (int) Math.floor( rect.getY() + ( rect.getHeight() - _currentPieceWidth ) / 2 );
		
		Point result = new Point( x1, y1 );

		return( result );
	}
	
	protected Rectangle getRectangleBoundsOfSquareByColRow( int col, int row )
	{
		int x1 = 0;
		int y1 = 0;
		int deltaX = 0;
		int deltaY = 0;

		if( (col>0) && (col<9) ) deltaX = _currentSquareWidth;
		else					 deltaX = _initialX_ofBoard;

		if( (row>0) && (row<9) ) deltaY = _currentSquareWidth;
		else					 deltaY = _initialX_ofBoard;

		if( _whitePlaysFromTheBottom )
		{
			if( col>0 ) x1 = (col-1)*_currentSquareWidth + _initialX_ofBoard;
			else		x1 = 0;

			if( row<9 ) y1 = (ChessBoard.NUM_OF_ROWS-row)*_currentSquareWidth + _initialX_ofBoard;
			else		y1 = 0;
		}
		else
		{
			if( col<9 ) x1 = (ChessBoard.NUM_OF_COLUMNS-col)*_currentSquareWidth + _initialX_ofBoard;
			else		x1 = 0;

			if( row>0 ) y1 = (row-1)*_currentSquareWidth + _initialX_ofBoard;
			else		y1 = 0;
		}
		
		Rectangle result = new Rectangle( x1, y1, deltaX, deltaY );
		return( result );
	}

	protected static int calculateSquareWidth( int width, int height )
	{
		int result = -1;

		if( (width>0) && (height>0) )
		{
			int squareSize1 = width / ( ChessBoard.NUM_OF_COLUMNS+1 );
			int squareSize2 = height / ( ChessBoard.NUM_OF_ROWS+1 );

			result = IntegerFunctions.min( squareSize1, squareSize2 );
		}

		result = IntegerFunctions.max( MIN_SQUARE_WIDTH, result );
		result = result & 0xFFFFFFFE; // we want an even width

		return( result );
	}

	protected void changeCurrentSquareWidth()
	{
		_currentSquareWidth = calculateSquareWidth( getWidth(), getHeight() );
		_currentPieceWidth = (int) Math.floor( _currentSquareWidth * _fractionOfPieceWidthOverSquareWidth );
		_initialX_ofBoard = _currentSquareWidth / 2;

/*
		try
		{
			_chessBoardImages.resizeSquaresAndPieces(_currentSquareWidth, _currentPieceWidth);
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
*/
	}

	@Override
	public void setSize( Dimension dim )
	{
		super.setSize( dim );

		changeCurrentSquareWidth();
	}

	@Override
	public void setSize( int width, int height )
	{
		super.setSize( width, height );
		
		changeCurrentSquareWidth();
	}
	
	@Override
	public void refresh()
	{
		repaint();
	}

	@Override
	public void pieceTopromotePawnChosen( String pieceTypeCode )
	{
		if( ( pieceTypeCode != null ) && (_promotionMove != null ) )
		{
			_promotionMove.setPromotionPiece( pieceTypeCode );
			makeNewMove( _promotionMove );
		}
		_promotionMove = null;
		_promotionJPanel.setVisible(false);
		repaint();
	}


	protected static ChessStringsConf getChessStrConf()
	{
		return( ChessStringsConf.instance() );
	}

	@Override
	public void changeLanguage(String newLanguage) throws Exception
	{
		refresh();
	}

	@Override
	public void unregisterFromChangeLanguageAsObserver()
	{
		if( _changeLanguageServer == null )
		{
			_changeLanguageServer.unregisterChangeLanguageObserver( this );
		}
	}

	@Override
	public void registerToChangeLanguageAsObserver( ChangeLanguageServerInterface conf)
	{
		if( _changeLanguageServer == null )
		{
			_changeLanguageServer = conf;
			_changeLanguageServer.registerChangeLanguageObserver( this );
		}
	}

	@Override
	public String getLanguage()
	{
		String result = null;
		if( _changeLanguageServer == null )
		{
			result = _changeLanguageServer.getLanguage();
		}
		return( result );
	}
}
