/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop;

import com.frojasg1.general.number.IntegerFunctions;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.font.FontRenderContext;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;

/**
 *
 * @author Usuario
 */
public class ImageFunctions
{
	protected static BufferedImage _internalStaticImage = null;
	protected static FontRenderContext _internalStaticFontRenderContext = null;
	
	static
	{
		_internalStaticImage = new BufferedImage( 10, 10, BufferedImage.TYPE_INT_RGB );

		Graphics gc = _internalStaticImage.getGraphics();
		_internalStaticFontRenderContext = ((Graphics2D)gc).getFontRenderContext();
	}

	public static BufferedImage resizeSquaredImage( BufferedImage original, int newWidth, Integer switchColorFrom,
													Integer switchColorTo, Integer alphaForPixelsDifferentFromColorFrom ) throws IllegalArgumentException
	{
		BufferedImage result = resizeImage( original, newWidth, newWidth,
											switchColorFrom, switchColorTo, alphaForPixelsDifferentFromColorFrom );

		return( result );
	}

	public static BufferedImage resizeImage( BufferedImage original, int width, int height, Integer switchColorFrom,
											Integer switchColorTo, Integer alphaForPixelsDifferentFromColorFrom ) throws IllegalArgumentException
	{
		int alpha = 0xFF000000;
		if( alphaForPixelsDifferentFromColorFrom != null )
			alpha = ( (alphaForPixelsDifferentFromColorFrom) & 0xFF ) << 24;

		if( ( width < 1 ) || ( height < 1 ) )		throw( new IllegalArgumentException( "Bad size for image.   Width: " + width + ". Height: " + height ) );

		BufferedImage result = new BufferedImage( width, height, BufferedImage.TYPE_INT_ARGB );

		double factorX = ((double) original.getWidth()) / width ;
		double factorY = ( (double) original.getHeight() ) / height;

		int[] pixels = getRGB( 0, 0, original.getWidth(), original.getHeight(), original );

		double transformedY = 0.5d;
		for( int tY = 0; tY < height; transformedY += 1, tY++ )
		{
			int originalOffsetY = (int) ( Math.floor(transformedY*factorY) ) * original.getWidth();
			double transformedX = 0.5D;
			for( int tX = 0; tX < width; transformedX += 1, tX++ )
			{
				int originalX = (int) ( Math.floor( transformedX*factorX ) );
				
				int pixelColor = pixels[ originalOffsetY+originalX ];
				
				if( ( switchColorFrom != null ) && ( pixelColor == switchColorFrom ) )
				{
					if( switchColorTo != null ) pixelColor = switchColorTo;
					else						pixelColor = pixelColor & 0xFFFFFF;
				}
				else if( alphaForPixelsDifferentFromColorFrom != null )
				{
					pixelColor = pixelColor & 0xFFFFFF | alpha;
				}

				result.setRGB( tX, tY, pixelColor );
			}
		}

		return( result );
	}


	/**
	* Returns an array of integer pixels in the default RGB color model
	* (TYPE_INT_ARGB) and default sRGB color space,
	* from a portion of the image data.  Color conversion takes
	* place if the default model does not match the image
	* <code>ColorModel</code>.  There are only 8-bits of precision for
	* each color component in the returned data when
	* using this method.  With a specified coordinate (x,&nbsp;y) in the
	* image, the ARGB pixel can be accessed in this way:
	* </p>
	*
	* <pre>
	*    pixel   = rgbArray[offset + (y-startY)*scansize + (x-startX)]; </pre>
	*
	* <p>
	*
	* An <code>ArrayOutOfBoundsException</code> may be thrown
	* if the region is not in bounds.
	* However, explicit bounds checking is not guaranteed.
	*
	* @param startX      the starting X coordinate
	* @param startY      the starting Y coordinate
	* @param w           width of region
	* @param h           height of region
	* @param rgbArray    if not <code>null</code>, the rgb pixels are
	*          written here
	* @param offset      offset into the <code>rgbArray</code>
	* @param scansize    scanline stride for the <code>rgbArray</code>
	* @return            array of RGB pixels.
	* @see #setRGB(int, int, int)
	* @see #setRGB(int, int, int, int, int[], int, int)
	*/
	public static int[] getRGB(int startX, int startY, int w, int h,
								BufferedImage bi )
	{
		ColorModel colorModel = bi.getColorModel();

		Raster raster = bi.getRaster();
//		WritableRaster raster = colorModel.createCompatibleWritableRaster( bi.getWidth(), bi.getHeight() );

		int scansize = w;
		int offset =0;
		
		int yoff  = offset;
		int off;
		Object data;
		int nbands = raster.getNumBands();
		int dataType = raster.getDataBuffer().getDataType();
		switch (dataType)
		{
			case DataBuffer.TYPE_BYTE:
			data = new byte[nbands];
				break;
			case DataBuffer.TYPE_USHORT:
				data = new short[nbands];
				break;
			case DataBuffer.TYPE_INT:
				data = new int[nbands];
				break;
			case DataBuffer.TYPE_FLOAT:
				data = new float[nbands];
				break;
			case DataBuffer.TYPE_DOUBLE:
				data = new double[nbands];
				break;
			default:
				throw new IllegalArgumentException("Unknown data buffer type: "+
													dataType);
		}

		int[] rgbArray = new int[offset+h*scansize];

		for (int y = startY; y < startY+h; y++, yoff+=scansize)
		{
			off = yoff;
			for (int x = startX; x < startX+w; x++)
			{
				if( (x>=0) && (x<bi.getWidth()) && (y>=0) && (y<bi.getHeight() ) )
				{
					rgbArray[off++] = colorModel.getRGB(raster.getDataElements(	x,
																				y,
																				data));
				}
				else
				{
					rgbArray[off++] = 0;
				}
		   }
	   }

		return rgbArray;
	}

	public static Rectangle2D getImageWrappedBoundsForString( Font font, String str )
	{
		Rectangle2D result = null;

		if( str != null )
		{
			FontRenderContext frc = _internalStaticFontRenderContext;
			result = font.getStringBounds( str, frc);
		}
		else
		{
			result = new Rectangle(0,0);
		}

		return( result );
	}

	public static void paintStringCentered( Graphics gc, Font font, String str, Color textColor, Rectangle bounds, Color backgroundColor )
	{
		if( backgroundColor != null )
		{
			gc.setColor( backgroundColor );
			gc.fillRect( (int) bounds.getX(), (int) bounds.getY(), (int) bounds.getWidth(), (int) bounds.getHeight() );
		}

		Point centralPoint = new Point( (int) bounds.getCenterX(), (int) bounds.getCenterY() );

		paintStringCentered( gc, font, str, textColor, centralPoint );
	}

	public static void paintStringCentered( Graphics gc, Font font, String str, Color textColor,
											Point centralPoint )
	{
		if( textColor != null ) gc.setColor( textColor );

		if( str != null )
		{
			FontRenderContext frc = ((Graphics2D)gc).getFontRenderContext();
			Rectangle2D wrappedBounds = font.getStringBounds(str, frc);
			int x1 = (int) Math.floor( centralPoint.getX() - wrappedBounds.getWidth() / 2 );
			int y1 = (int) Math.floor( centralPoint.getY() + wrappedBounds.getHeight() / 2 ) - 3;

			gc.setFont( font );
			gc.drawString( str, x1, y1);
		}
	}


	public static void paintClippedImage( Component comp, Graphics gc, BufferedImage bi, Point leftUpperCorner )
	{
		if( bi != null )
		{
			// if some part of the piece is inside the visible zone (the screen)
			if( ( leftUpperCorner.getX() < comp.getWidth() ) && ( leftUpperCorner.getX() + bi.getWidth() >= 0 ) &&
				( leftUpperCorner.getY() < comp.getHeight() ) && ( leftUpperCorner.getY() + bi.getHeight() >= 0 ) )
			{
				int dx1 = IntegerFunctions.max( 0, (int) leftUpperCorner.getX() );
				int dy1 = IntegerFunctions.max( 0, (int) leftUpperCorner.getY() );
				int dx2 = IntegerFunctions.min( comp.getWidth(), (int) ( leftUpperCorner.getX() + bi.getWidth() ) );
				int dy2 = IntegerFunctions.min( comp.getHeight(), (int) ( leftUpperCorner.getY() + bi.getHeight() ) );

				int sx1 = dx1 - (int) leftUpperCorner.getX();
				int sy1 = dy1 - (int) leftUpperCorner.getY();
				int sx2 = sx1 + dx2 - dx1;
				int sy2 = sy1 + dy2 - dy1;

				gc.drawImage( bi, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, null );
			}
		}
	}

	public static void drawRect( Graphics gc, int xx, int yy, int width, int height, Color color, int thick )
	{
		gc.setColor( color );

		for( int ii=0; ii<thick; ii++ )
		{
			gc.drawRect( xx - ii, yy - ii,	width + 2 * ii, height + 2 * ii );
		}
	}

}
